# -*- coding: utf-8 -*-
import asyncio
import subprocess
import sys
import time
from pathlib import Path
from typing import Dict
from typing import List
from typing import Tuple

import typer
import yaml
from blessed import Terminal
from cs_sync.github import chain
from cs_sync.github import group
from cs_sync.handlers import parse_repo
from cs_sync.helpers import expand_path
from cs_sync.helpers import flatten_list


term = Terminal()
cli = typer.Typer()


# Config
config_file = Path.home() / '.cssync'


def load_config() -> Tuple[List, List]:
    """Get configuration from config file.

    Returns repo_paths and bare_repo_dicts.
    """
    if config_file.exists():
        with open(config_file, "r") as ymlfile:
            config = yaml.load(ymlfile, Loader=yaml.Loader)
        repo_paths = flatten_list(
            [expand_path(i) for i in config.get('repo_paths', [])]
        )
        bare_repo_dicts: List[Dict] = config.get('bare_repos', [])
        bare_repo: Dict[str, str]
        for bare_repo in bare_repo_dicts:
            bare_repo['git_dir'] = expand_path(bare_repo['git_dir'])[0]
            bare_repo['work_tree'] = expand_path(bare_repo['work_tree'])[0]
    else:
        repo_paths = []
        bare_repo_dicts = []
    return repo_paths, bare_repo_dicts


def measure_time_async(original_async_function):
    async def wrapper(*args, **kwargs):

        # Work Before
        start = time.perf_counter()

        # Run Async Function
        results = await original_async_function(*args, **kwargs)

        # Work After
        elapsed = time.perf_counter() - start
        print(term.red(f"Executed in {elapsed:0.2f} seconds."))

        # Return results
        return results

    return wrapper


def measure_time(original_async_function):
    def wrapper(*args, **kwargs):

        # Work Before
        start = __import__('time').perf_counter()

        # Run Async Function
        results = original_async_function(*args, **kwargs)

        # Work After
        elapsed = __import__('time').perf_counter() - start
        print(term.red(f"Executed in {elapsed:0.2f} seconds."))

        # Return results
        return results

    return wrapper


async def chain_handler(async_def, handler, *args, **kwargs):
    results = await async_def
    await handler(results, *args, **kwargs)
    return results


@cli.command()
def all(short: bool = typer.Option(False, "--short")):
    """Git, AWS, System Settings (Windows Terminal), etc."""
    git(short=short)
    task()


@measure_time
@cli.command()
def git(short: bool = typer.Option(False, "--short")):
    """Status, Pull, etc. all git repos."""

    repo_paths, bare_repo_dicts = load_config()
    repo_paths = [
        i
        for i in repo_paths
        if Path(i).is_dir() and '.git' in [j.name for j in Path(i).glob('*')]
    ]

    repos = repo_paths + bare_repo_dicts
    chains = [chain_handler(chain(r), parse_repo, short) for r in repos]
    tasks = group(chains)
    if sys.version > '3.6':
        _ = asyncio.run(tasks)
    else:
        loop = asyncio.get_event_loop()
        _ = loop.run_until_complete(tasks)


@cli.command()
def task():
    """Sync Taskwarrior with Taskserver."""

    command = ['task', 'sync']
    if sys.platform == 'win32':
        command = ['wsl'] + command
    subprocess.run(command)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
import re
from types import SimpleNamespace

from blessed import Terminal

term = Terminal()


def repo_output_handler(repo, tabs=1, indent='\t'):

    results = []
    repo_condition = 'green'
    for i in repo.modified:
        results.append(term.yellow(f'{tabs*indent}modified: {i.path}'))
        repo_condition = 'yellow'
    for i in repo.renamed:
        results.append(
            term.yellow(f'{tabs*indent}renamed: {i.path} {i.original_path}')
        )
        repo_condition = 'yellow'
    for i in repo.untracked:
        results.append(term.yellow(f'{tabs*indent}untracked: {i.path}'))
        repo_condition = 'yellow'
    for i in repo.deleted:
        results.append(term.red(f'{tabs*indent}deleted: {i.path}'))
        repo_condition = 'red'
    for i in repo.ignored:
        results.append(term.red(f'{tabs*indent}ignored: {i.path}'))

    if not repo.online:
        repo_condition = 'blue'

    try:
        header = [(tabs - 1) * indent, repo.name]
        if repo.branch.head:
            header.append(term.magenta(repo.branch.head))
        if repo.ahead:
            header.append(term.cyan(f'↑{repo.ahead}'))
        if repo.behind:
            header.append(term.cyan(f'↓{repo.ahead}'))
        if repo.modified:
            header.append(term.cyan(f'~{len(repo.modified)}'))
        if repo.deleted:
            header.append(term.cyan(f'-{len(repo.deleted)}'))
        if repo.untracked:
            header.append(term.cyan(f'…{len(repo.untracked)}'))
        header = ' '.join(header)
    except Exception:  # For Bare Repos
        header = f"{(tabs-1)*indent} {repo.name} {len(repo.modified):+}"

    conditions = {
        "green": lambda iterable, item: iterable.insert(
            0, term.green(f"✓ {item}")
        ),
        "yellow": lambda iterable, item: iterable.insert(
            0, term.yellow(f"⚠ {header}")
        ),
        "red": lambda iterable, item: iterable.insert(
            0, term.red(f"! {header}")
        ),
        "blue": lambda iterable, item: iterable.insert(
            0, term.blue(f"? {header}")
        ),
        "default": lambda iterable, item: iterable.insert(0, f"? {header}"),
    }
    add_header = conditions.get(repo_condition, conditions['default'])
    add_header(results, header)

    return '\n'.join(results)


async def parse_repo(parsed_output: dict, short: bool = False):
    status_stdout = parsed_output['status']['stdout']
    parsed = parse_git_status(status_stdout)
    parsed.name = parsed_output['name']

    if short:
        # To tell if there's any changes that were made.
        any_changes = any(
            i
            for i in [
                parsed.ahead,
                parsed.behind,
                len(parsed.modified),
                len(parsed.renamed),
                len(parsed.deleted),
                len(parsed.untracked),
                len(parsed.ignored),
            ]
        )
        if any_changes:
            print(repo_output_handler(parsed))
    else:
        print(repo_output_handler(parsed))


def parse_git_status(stdout):

    lines = stdout.splitlines()
    repo = SimpleNamespace()

    branch_info = [i for i in lines if i.startswith('#')]
    modified = [i for i in lines if i.startswith('1')]
    renamed_or_copied = [i for i in lines if i.startswith('2')]
    untracked = [i for i in lines if i.startswith('?')]
    ignored = [i for i in lines if i.startswith('!')]

    # Branch
    oid_group = '# branch.oid (?P<oid>.*)'
    head_group = '# branch.head (?P<head>.*)'
    upstream_group = '# branch.upstream (?P<upstream>.*)'
    ahead_behind_group = '# branch.ab (?P<ahead>.*) (?P<behind>.*)'
    space = r'\s*'
    branch_re = re.compile(
        fr'({oid_group})?'
        + fr'{space}'
        + fr'({head_group})?'
        + fr'{space}'
        + fr'({upstream_group})?'
        + fr'{space}'
        + fr'({ahead_behind_group})?'
    )

    branch_info = [i for i in lines if i.startswith('#')]
    branch_match = branch_re.match('\n'.join(branch_info))

    branch = SimpleNamespace(
        oid=branch_match.group('oid'),
        head=branch_match.group('head'),
        upstream=branch_match.group('upstream'),
        ahead=int(
            branch_match.group('ahead') if branch_match.group('ahead') else 0
        ),
        behind=int(
            branch_match.group('behind') if branch_match.group('behind') else 0
        ),
    )

    # Changed
    modified = [get_file_info(i.split(maxsplit=9)) for i in modified]

    # Renamed or Copied
    renamed_or_copied = [
        get_file_info(i.split(maxsplit=10)) for i in renamed_or_copied
    ]

    # Untracked
    untracked = [i.split(maxsplit=1)[1] for i in untracked]
    untracked = [SimpleNamespace(path=i, type='Untracked') for i in untracked]

    # Ignored
    # Only if `--ignored=matching` is included
    ignored = [i.split(maxsplit=1)[1] for i in ignored]
    ignored = [SimpleNamespace(path=i, type='Ignored') for i in ignored]

    # All Files
    all_files = modified + renamed_or_copied + untracked + ignored
    # Resort by Type
    modified = [i for i in all_files if i.type[0] == 'M']
    renamed = [i for i in all_files if i.type[0] == 'R']
    deleted = [i for i in all_files if i.type[0] == 'D']
    untracked = [i for i in all_files if i.type[0] == 'U']
    ignored = [i for i in all_files if i.type[0] == 'I']

    repo.branch = branch
    repo.ahead = branch.ahead
    repo.behind = branch.behind
    repo.modified = modified
    repo.renamed = renamed
    repo.deleted = deleted
    repo.untracked = untracked
    repo.ignored = ignored
    repo.all_changed_files = all_files

    # Quickfix
    if repo.branch.oid:
        repo.online = True
    else:
        repo.online = False

    return repo


def get_file_info(raw):

    if raw[0] == '1':
        type_ = 'changed'
    else:
        type_ = 'renamed_or_copied'
    raw = raw[1:]  # Get rid of the type as the docs don't refer to it.

    if type_ == 'renamed_or_copied':
        path = raw[8]
        original_path = raw[9]
    else:
        path = raw[7]
        original_path = None

    staged = False
    subtype = raw[0][-1] if raw[0][-1] != '.' else raw[0][0]
    if raw[0][0] == '.':
        staged = False
        subtype = raw[0][-1]
    elif raw[0][-1] == '.':
        staged = True
        subtype = raw[0][0]
    if subtype == 'D':
        subtype = 'Deleted'
    elif subtype == 'M':
        subtype = 'Modified'
    elif subtype == 'R':
        subtype = 'Renamed'

    file = SimpleNamespace(
        path=path,
        staged=staged,
        original_path=original_path,
        type=subtype,
    )

    return file



# -*- coding: utf-8 -*-
import asyncio
from pathlib import Path
from typing import Coroutine
from typing import Iterable


async def async_run_command(command):
    """Run an async command and return stdout and stderr."""
    process = await asyncio.create_subprocess_shell(
        command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await process.communicate()
    stdout, stderr = stdout.decode(), stderr.decode()

    return stdout, stderr


async def async_git_pull(repo_path=None, git_dir=None, work_tree=None):
    """`git pull` on a directory, or git directory and work tree for bare
    repos."""
    if repo_path:
        command = ['git', '-C', fr'"{repo_path}"', 'pull']
        command = ' '.join(command)
    else:
        command = [
            'git',
            f'--git-dir="{git_dir}"',
            f'--work-tree="{work_tree}"',
            'pull',
        ]
        command = ' '.join(command)

    stdout, stderr = await async_run_command(command)

    return stdout, stderr


async def async_git_push(repo_path=None, git_dir=None, work_tree=None):
    """`git push` on a directory, or git directory and work tree for bare
    repos."""
    if repo_path:
        command = ['git', '-C', fr'"{repo_path}"', 'push']
        command = ' '.join(command)
    else:
        command = [
            'git',
            f'--git-dir="{git_dir}"',
            f'--work-tree="{work_tree}"',
            'push',
        ]
        command = ' '.join(command)

    stdout, stderr = await async_run_command(command)

    return stdout, stderr


async def async_git_status(repo_path=None, git_dir=None, work_tree=None):
    """`git status` on a directory, or git directory and work tree for bare
    repos."""

    if repo_path:
        command = [
            'git',
            '-C',
            fr'"{ repo_path }"',
            'status',
            '--porcelain=2',
            '-b',
        ]
        command = " ".join(command)
    else:
        command = [
            'git',
            f'--git-dir="{git_dir}"',
            f'--work-tree="{work_tree}"',
            'status',
            '--porcelain=2',
            '-b',
        ]
        command = ' '.join(command)
    stdout, stderr = await async_run_command(command)

    return stdout, stderr


async def chain(repo_path=None, git_dir=None, work_tree=None, name=None):
    """Chain pull, status, and push (for bare repos)."""

    if type(repo_path) != str:

        git_dir = repo_path['git_dir']
        work_tree = repo_path['work_tree']
        name = repo_path.get('name', None)
        repo_path = None

    if name:
        pass
    elif repo_path:
        name = Path(repo_path).name
    else:
        name = Path(git_dir).name

    results = dict(name=name, pull={}, status={}, push={})

    if not repo_path:  # Bare Repos
        # Can't check ahead-behind, so...
        stdout, stderr = await async_git_push(repo_path, git_dir, work_tree)
        results['push']['stdout'], results['push']['stderr'] = stdout, stderr
    stdout, stderr = await async_git_pull(repo_path, git_dir, work_tree)
    results['pull']['stdout'], results['pull']['stderr'] = stdout, stderr
    stdout, stderr = await async_git_status(repo_path, git_dir, work_tree)
    results['status']['stdout'], results['status']['stderr'] = stdout, stderr

    return results


async def group(repos: Iterable[Coroutine]):
    """Main function, but async.
    Chain pulls and status checks to each repo/bare repo, then run them.
    """
    return await asyncio.gather(*repos)



# -*- coding: utf-8 -*-
from glob import glob
from pathlib import Path
from typing import List
from typing import Union


def expand_path(path: Union[str, Path]) -> List[str]:
    paths: List[str] = glob(str(Path(path).expanduser()))
    return paths


def flatten_list(nested_list: Union[list, tuple]) -> list:

    supported_types = (list, tuple)

    results = []
    for item in nested_list:
        if not isinstance(item, supported_types):
            results.append(item)
        else:
            results.extend(flatten_list(item))

    return results
